#include <t3devlib/t3devlib.h>

using namespace t3devlib;


// 6lowpan Platform Adapter implementation
//
// Please read the "Implementing Adapters" section of the T3DevKit
// documentation in order to understand how this PA is designed.
//
// --->  http://t3devkit.gforge.inria.fr/doc/userref/
//


// @desc	external function fx_payloadLength (defined in LibIpv6_ExternalFunction)
//
// @param 0	an IPv6 packet
//
// @return	the length of the encoded payload of this IPv6 packet

bool fx_payloadLength (ParameterList& par_list, Bitstring& ret_val) {
	
	ret_val.SetValueBin (Unsigned (16, (par_list[0].GetLength()/8) - 40));

	return true;
}

// @desc	external function fx_encodeMessage (defined in LibIpv6_ExternalFunction)
//
// @param 0	an IPv6 packet
//
// @return	the same IPv6 packet (encoded)

bool fx_encodeMessage (ParameterList& par_list, Bitstring& ret_val) {
	
	ret_val.SetValueBin (par_list[0]);

	return true;
}


// @desc	Compute an IPv6 checksum
//		This function assembles an IPv6 pseudo-header and computes the
//		checksum of the packet.
//
// @remark	 The checksum field must be set to zero prior to calling this function.
//
//
// @param bs_ip6		whole Datagram (including Ipv6 option headers)
// @param checksum_result	a variable that will hold the resulting checksum
// 				(in network order)
//
// @param next_header		a variable that will receive the next header value of the pseudo-header
//
// @return			true if successful

static bool compute_ip6_checksum (const Bitstring& bs_ip6, uint16_t& checksum_result, uint8_t& next_header)
{

	const uint8_t* ip6;

	const uint8_t* ip6_src;
	const uint8_t* ip6_dst;

	const uint8_t* payload;
	uint32_t length;

	unsigned int i;

	/* IPv6 Header */
	ip6 = (const uint8_t*) bs_ip6.GetValueBin();

	ip6_src = ip6 + 8;
	ip6_dst = ip6 + 24;

	next_header = *(uint8_t*) (ip6 + 6);

	payload = ip6 + 40;
	length  = ntohs (*(uint16_t*) (ip6 + 4));

	if (length + 40 != bs_ip6.GetLength()/8)
	{
		cerr << "PA Error: cannot compute IPv6 checksum (payload length in ip6 header does not match the size of the packet)" << endl;
		return false;
	}

	// assemble the extension header
	char pshdr[40];
	memcpy (pshdr,      ip6_src, 16);
	memcpy (pshdr + 16, ip6_dst, 16);
	*(uint32_t*)(pshdr + 32) = htonl (length);
	*(uint32_t*)(pshdr + 36) = 0;
	*(uint8_t *)(pshdr + 39) = next_header;

	// compute the checksum
	uint32_t checksum = 0;
	const uint16_t* ptr;

	//	pseudo header
	ptr = (uint16_t *) (pshdr);
	for(i=0 ; i<20 ; i++)
	{
		checksum += htons (ptr[i]);
		checksum = (checksum >> 16) + (checksum & 0x0000ffff);
	}

	//	payload
      
	ptr = (uint16_t *) payload;
	for(i=0; i<length/2 ; i++)
	{
	    checksum += htons (ptr[i]);  
	    checksum = (checksum >> 16) + (checksum & 0x0000ffff);
	}
  
  	//	last byte (if payload length is odd)
	if (length % 2)
	{  
		uint16_t val = ((uint8_t *) ptr) [length - 1];
		val <<= 8;

		checksum += val;  
		checksum = (checksum >> 16) + (checksum & 0x0000ffff);
	}

	checksum = (~ checksum) & 0x0000ffff;

	checksum_result = htons (checksum);

	return true;
}

// @desc	external function fx_icmpv6Checksum (defined in LibIpv6_ExternalFunction)
//
// @param 0	an IPv6 packet containing an ICMPv6 message (note: the icmpv6
//		checksum field must be set to 0)
//
// @return	the value of the computed checksum (in network order)
//
bool fx_icmpv6Checksum (ParameterList& par_list, Bitstring& ret_val)
{
	uint8_t  next_header;
	uint16_t checksum;

	if (!compute_ip6_checksum (par_list[0], checksum, next_header))
		return false;
	
	if (next_header != 58) {
		cerr << "Error: fx_icmpv6Checksum: expected next header 58, got " << next_header << endl;
		return false;
	}

	ret_val.SetValueBin ((uint8_t*)&checksum, 16);

	return true;
}



//////////////////////////////////////////////////////////////////////////////
// compression & decompression for 6lowpan-nhc                              //
//////////////////////////////////////////////////////////////////////////////


#define NEXT_HEADER_UDP		17
#define NEXT_HEADER_IPV6	41

namespace t3devlib
{
	boost::recursive_mutex::scoped_lock getCodecLock();
}

// abstract base class for delayed operations handled by DelayedLifo
class DelayedOperation
{
public:
	virtual bool operator() (Buffer& buff) = 0;
};


// Internals for DelayedLIFO
typedef std::vector< boost::shared_ptr<DelayedOperation> >  DelayedLIFOBase;

// A class for handling delayed operations in the codec
//
// This class is used for delaying some operations during the compression or
// decompression of a 6lowpan frame. For example: decompressing an IPv6 packet
// requires to compute the payload length field from the decompressed packet,
// however this can be done only once the packet is fully decompressed.
//
// New delayed operations are added using the push_back() function. Then they
// are processed in the LIFO order.
//
class DelayedLIFO : private DelayedLIFOBase
{
public:
	using DelayedLIFOBase::empty;
	using DelayedLIFOBase::back;
	using DelayedLIFOBase::push_back;
	using DelayedLIFOBase::pop_back;

	// push a new delayed operation
	void push_back (DelayedOperation* op)
	{
		push_back (boost::shared_ptr<DelayedOperation> (op));
	}
};


// @desc	Use the 6lowpan CoDec to decode a binary string into a TTCN-3 message.
//
// @param	buff	binary buffer to be decoded
// @param	output	the variable that will hold the result (must inherit
//			from t3devlib::Variable)
//
// @return		true if successfut
template <class T>
static bool decode (Buffer& buff, T& output)
{
	try {
		boost::recursive_mutex::scoped_lock l (t3devlib::getCodecLock());
		output.Decode (buff);
		return true;
	} catch (DecodeError& ex) {
		cerr << "unable to decode " << output.GetTypeName() << " message" << endl << ex;
		output.Dump (cerr);
		buff.Dump (cerr);
		return false;
	}
}

// @desc	Use the 6lowpan CoDec to decode a binary string into a TTCN-3 message.
//
// @param	input	binary message to be decoded
// @param	output	the variable that will hold the result (must inherit
//			from t3devlib::Variable)
//
// @return		true if successfut
template <class T>
static bool decode (Bitstring& input, T& output)
{
	Buffer buff;
	buff.SetValueBin (input);
	return decode (buff, output);
}

// @desc	Use the 6lowpan CoDec to encode a TTCN-3 message into a binary string.
//
// @param	input	the TTCN-3 message to be encoded (must inherit from
//			t3devlib::Variable)
// @param	buff	binary buffer that will hold the result
//
// @return		true if successfut
template <class T>
static bool encode (T& input, Buffer& buff)
{
	try {
		boost::recursive_mutex::scoped_lock l (t3devlib::getCodecLock());
		input.Encode (buff);
		return true;
	} catch (DecodeError& ex) {
		cerr << "unable to encode " << input.GetTypeName() << " message" << endl << ex;
		input.Dump (cerr);
		buff.Dump (cerr);
		return false;
	}
}

// @desc	Use the 6lowpan CoDec to encode a TTCN-3 message into a binary string.
//
// @param	input	the TTCN-3 message to be encoded (must inherit from
//			t3devlib::Variable)
// @param	output	bitstring that will hold the result
//
// @return		true if successfut
template <class T>
static bool encode (T& input, Bitstring& output)
{
	Buffer buff;
	if (!encode (input, buff))
		return false;

	buff.SetPosition (0);
	output.SetValueBin (buff.GetValueBin(), buff.GetLength());
	return true;
}

#include "gen_classes.h"
using namespace t3devlib::gen;

// @desc	constant holding the IPv6 unspecified address
// 		-> 0000:0000:0000:0000:0000:0000:0000:0000
static const uint32_t IPV6_UNSPECIFIED_ADDRESS[4] = {0, 0, 0, 0};

// @desc	constant holding the 6 first bytes of an IPv6 interface
//		identifier derived from a IEEE 802.15.4 short address
//		-> 0000:00ff:fe00
static const uint8_t  SHORT_ADDR_IID_PATTERN[6] = {0, 0, 0, 0xff, 0xfe, 0};

// @desc	constant holding the IPv6 link-local prefix
// 		-> fe80:0000:0000:0000
static const uint8_t  IPV6_LL_PREFIX[8] = {0xfe, 0x80, 0, 0, 0, 0, 0, 0};


// @desc	compress an IPv6 unicast address
//
// @param	ipv6_address	IPv6 address to be compressed
// @param	hw		hardware address of the lower layer
//
// @param	am		resulting AM value
// @param	ac		resulting AC value
// @param	inl		resulting Inline value
//
// @param	compress_unspec	true if :: can be compressed to AM:00, AC:1
// 				(this is for source address only)
static
void compress_unicast (	const Ipv6Address& ipv6_address,
			const SixLowpanHwAddress& hw,
			UInt2& am,
			Boolean& ac,
			Octetstring& inl,
			bool compress_unspec)
{
	const uint8_t* ipv6 = ipv6_address.GetValueBin();

	if (compress_unspec && memcmp (ipv6, IPV6_UNSPECIFIED_ADDRESS, 16) == 0) {
		// unspecified address
		ac.SetValue (true);
		am.SetValue (0);
		return;
	}

	// interface id
	if ((hw.GetLength() == 64) && (ipv6[8]^2 == hw[0])&& (memcmp (ipv6+9, hw.GetValueBin()+1, 7) == 0))
	{
		// full compression (EUI 64)
		am.SetValue (3);
	
	} else if (memcmp (ipv6+8, SHORT_ADDR_IID_PATTERN, 6) == 0) {
		// short address pattern
		if ((hw.GetLength() == 16) && (memcmp (ipv6+14, hw.GetValueBin(), 2) == 0)) {
			// full compression (short)
			am.SetValue (3);
		} else {
			// compression to 16 bits
			am.SetValue (2);
		}
	} else {
		// no iid compression
		am.SetValue (1);
	}

	// prefix
	if (memcmp (ipv6, IPV6_LL_PREFIX, 8) == 0) {
		// link-local
		ac.SetValue (0);
	} else {
		// no compression  (stateful compression not supported)
		ac.SetValue (0);
		am.SetValue (0);
	}

	// fill inline address
	switch (am.GetValue())
	{
	case 0:	// 128 bits
		inl.SetValueBin (ipv6, 128);
		break;
	case 1:	// 64 bits
		inl.SetValueBin (ipv6+8, 64);
		break;
	case 2:	// 16 bits
		inl.SetValueBin (ipv6+14, 16);
		break;
	case 3:	// 0 bits
		break;
	}
}

// @desc	decompress an IPv6 unicast address
//
// @param	ipv6_address	resulting IPv6 address
//
// @param	hw		hardware address of the lower layer
// @param	am		AM value
// @param	ac		AC value
// @param	inl		Inline value
static
void decompress_unicast (Ipv6Address& ipv6_address,
			const SixLowpanHwAddress& hw,
			const UInt2& am_field,
			const Boolean& ac_field,
			const Octetstring& inl)
{
	uint8_t ac = ac_field.GetValue();
	uint8_t am = am_field.GetValue();
	if (am == 0)
	{
		if (ac == 0) {
			// in-line address
			ipv6_address.SetValueBin (inl);
		} else {
			// unspecified address
			ipv6_address.SetValueBin ((uint8_t*) IPV6_UNSPECIFIED_ADDRESS, 128);
		}
		return;
	}

	uint8_t buff[16];
	
	// Stateless decompression only (fe80::)
	memcpy (buff, IPV6_LL_PREFIX, 8);

	switch (am)
	{
	case 1:	// 64 bits
		memcpy (buff+8, inl.GetValueBin(), 8);
		break;
	
	case 2: // 16 bits
		memcpy (buff+8, SHORT_ADDR_IID_PATTERN, 6);
		memcpy (buff+14, inl.GetValueBin(), 2);
		break;
	
	case 3: // derived from the hw address
		switch (hw.GetLength())
		{
		case 16: // short address
			memcpy (buff+8, SHORT_ADDR_IID_PATTERN, 6);
			memcpy (buff+14, hw.GetValueBin(), 2);
			break;
		case 64: // EUI 64
			memcpy (buff+8, hw.GetValueBin(), 8);
			// invert u/l bit
			buff[8] ^= 2;
			break;
		default:
			cerr << "decompress_unicast: don't know how to handle "
			     << hw.GetLength() << "-bit hardware addresses" << endl;
			break;
		}
	}
	ipv6_address.SetValueBin (buff, 128);
}

// @desc	check if the next header can be compressed
//
// @param	next_header	IPv6 Next Header
//
// @return	true if this implementation knows how to compress this kind of
//		next header
static bool can_be_compressed (uint8_t next_header)
{
	switch (next_header)
	{
	case NEXT_HEADER_IPV6:
	case NEXT_HEADER_UDP:
		return true;
	default:
		return false;
	}
}

// @desc	compress an IPv6 header
//
// @param buff		buffer containing the uncompressed IPv6 header
// @param next_header	variable that will be filled with
// 			- -1 if the next header cannot be compressed
// 			- the value of the NextHeader (identifying the type of
// 			  the next header) otherwise
// @param frame		reference to 6lowpan frame containing this IPv6 packet
//
// @return		a pointer to a new SixLowpanIPHC object (containing the
//			compressed header)
static Variable* compress_iphc (Buffer& buff, int& next_header, SixLowpanFrame& frame)
{
	// read the IPv6 header
	Ipv6Header ip6;
	if (!decode (buff, ip6))
		return NULL;
	
	// compress the IPv6 header
	auto_ptr<SixLowpanIPHC> iphc (new SixLowpanIPHC);

	// dispatch value
	iphc->Get_dispatch().SetValue("\x60", 3);

	// ECN + DSCP + Flow Label
	int fl = ip6.Get_flowLabel().GetValue();
	uint8_t tc = ip6.Get_trafficClass().GetValue();
	uint8_t dscp = tc & 0x3f;
	uint8_t ecn  = (tc & 0xc0) >> 6;
	uint8_t tf;
	if (fl) {
		if (dscp)
			// all fields present
			tf = 0;
		else
			// DSCP elided
			tf = 1;
	} else {
		if (tc)
			// flow label elided
			tf = 2;
		else
			// all fields elided
			tf = 3;
	}
	
	iphc->Get_tf().SetValue (tf);
	if (tf < 3)
		iphc->Get_inline_ecn().SetValue (ecn);
	if ((tf % 2) == 0)
		iphc->Get_inline_dscp().SetValue (dscp);
	if (tf < 2) {
		iphc->Get_inline_tf_pad().SetValueBin ("", (tf ? 2 : 4));
		iphc->Get_inline_fl().SetValue (fl);
	}

	// Hop Limit
	int hl;
	switch (ip6.Get_hopLimit().GetValue())
	{
	case 1:		hl = 1;		break;
	case 64:	hl = 2;		break;
	case 255:	hl = 3;		break;
	default:	hl = 0;		break;
	}

	iphc->Get_hl().SetValue (hl);
	if (hl == 0)
		iphc->Get_inline_hl().SetValue (ip6.Get_hopLimit());

	// Context ID (not supported)
	iphc->Get_cid().SetValue(false);
	
	// Source Address
	compress_unicast (	ip6.Get_sourceAddress(),
				frame.Get_hwSrc(),
				iphc->Get_sam(),
				iphc->Get_sac(),
				iphc->Get_inline_src(),
				true);

	// Destination Address
	const uint8_t* ip6_dst = ip6.Get_destinationAddress().GetValueBin();
	if (ip6_dst[0] != 0xff) {
		// unicast address
		iphc->Get_m().SetValue(0);

		compress_unicast (	ip6.Get_destinationAddress(),
					frame.Get_hwDst(),
					iphc->Get_dam(),
					iphc->Get_dac(),
					iphc->Get_inline_dst(),
					false);
	} else {
		// multicast address
		iphc->Get_m().SetValue(1);

		iphc->Get_dac().SetValue(0); // stateless compression only

		uint8_t dam;
		if ((ip6_dst[1] == 2) && (memcmp (ip6_dst+2, IPV6_UNSPECIFIED_ADDRESS, 13) == 0)) {
			// 8 bits
			dam = 3;
		} else if (memcmp (ip6_dst+2, IPV6_UNSPECIFIED_ADDRESS, 11) == 0) {
			// 32 bits
			dam = 2;
		} else if (memcmp (ip6_dst+2, IPV6_UNSPECIFIED_ADDRESS, 9) == 0) {
			// 48 bits
			dam = 1;
		} else {
			// 128 bits
			dam = 0;
		}
		iphc->Get_dam().SetValue (dam);

		// fill the inline field
		uint8_t b[6];
		switch (dam) {
		case 0:	// 128 bits
			iphc->Get_inline_dst().SetValueBin (ip6_dst, 128);
			break;
		case 1:	// 48 bits
			b[0]=ip6_dst[1];
			memcpy (b+1, ip6_dst+11, 5);
			iphc->Get_inline_dst().SetValueBin (b, 48);
			break;
		case 2:	// 32 bits
			b[0]=ip6_dst[1];
			memcpy (b+1, ip6_dst+13, 3);
			iphc->Get_inline_dst().SetValueBin (b, 32);
			break;
		case 3:	// 8 bits
			iphc->Get_inline_dst().SetValueBin (ip6_dst+15, 8);
			break;
		}
	}

	// set isrc and idst to omit in case they are empty
	if (!iphc->Get_inline_src().GetLength())
		iphc->SetField (iphc->id_inline_src, new Undef());
	if (!iphc->Get_inline_dst().GetLength())
		iphc->SetField (iphc->id_inline_dst, new Undef());

	
	// Next Header
	next_header = ip6.Get_nextHeader().GetValue();
	if (can_be_compressed (next_header))
	{
		iphc->Get_nh().SetValue(1);
	} else {
		iphc->Get_nh().SetValue(0);
		iphc->Get_inline_nh().SetValue (ip6.Get_nextHeader());
		next_header = -1;
	}

	return iphc.release();
}


// @desc	compress a UDP header
//
// @param buff		buffer containing the uncompressed UDP
// @param next_header	variable that will be filled with -1
// 			(there is no IPv6 NextHeader after UDP) 
//
// @return		a pointer to a new SixLowpanNHCUDP object (containing
//			the compressed header)
static Variable* compress_nhc_udp (Buffer& buff, int& next_header)
{
	// no further compression
	next_header = -1;

	// decode the header
	UInt16 u16;
	uint16_t sport, dport, checksum;

	u16.Decode (buff);
	sport = u16.GetValue();

	u16.Decode (buff);
	dport = u16.GetValue();

	u16.Decode (buff);
	// payload length (ignored)

	u16.Decode (buff);
	checksum = u16.GetValue();


	auto_ptr<SixLowpanNHCUDP> uhc (new SixLowpanNHCUDP);

	uhc->Get_dispatch().SetValueBin ("\xf0", 5);

	// checksum
	uhc->Get_c().SetValue (0);
	uhc->Get_inline_checksum().SetValue (checksum);

	// ports
	if (((sport & 0xfff0) == 0xf0b0) &&
	    ((dport & 0xfff0) == 0xf0b0))
	{
		// 4/4
		uhc->Get_p().SetValue (3);

		uint8_t c;
		c = (sport & 0xf) << 4;
		uhc->Get_inline_sport().SetValueBin (&c, 4);

		c = (dport & 0xf) << 4;
		uhc->Get_inline_dport().SetValueBin (&c, 4);
	}
	else if ((sport & 0xff00) == 0xf000)
	{
		// 8/16
		uhc->Get_p().SetValue (2);

		uint8_t c;
		c = sport & 0xff;
		uhc->Get_inline_sport().SetValueBin (&c, 8);
	
		UInt16 w;
		w.SetValue (dport);
		uhc->Get_inline_dport().SetValueBin (w);
	}
	else if ((dport & 0xff00) == 0xf000)
	{
		// 16/8
		uhc->Get_p().SetValue (1);
	
		UInt16 w;
		w.SetValue (sport);
		uhc->Get_inline_sport().SetValueBin (w);

		uint8_t c;
		c = dport & 0xff;
		uhc->Get_inline_dport().SetValueBin (&c, 8);
	}
	else
	{
		// 16/16
		uhc->Get_p().SetValue (0);
	
		UInt16 w;
		w.SetValue (sport);
		uhc->Get_inline_sport().SetValueBin (w);

		w.SetValue (dport);
		uhc->Get_inline_dport().SetValueBin (w);
	}

	return uhc.release();
}

// @desc	external function fx_compress6lowpan (defined in Ats6lowpan)
//
// @param 0	an uncompressed SixLowpanFrame
//
// @return	a compressed SixLowpanFrame
bool fx_compress6lowpan (ParameterList& par_list, Bitstring& ret_val)
{
	SixLowpanFrame frame;
	if (!decode (const_cast<Bitstring&> (par_list[0]), frame))
		return false;

	if ((!frame.IsPresent (frame.id_dispatch)) ||
	    (!frame.IsPresent (frame.id_ipv6)))
	{
		cerr << "fx_compress6lowpan: frame does not contain an ipv6 packet" << endl;
		return false;
	}

	if (frame.IsPresent (frame.id_hc))
	{
		cerr << "fx_compress6lowpan: frame already has a hc field" << endl;
		return false;
	}

	if (frame.IsPresent (frame.id_payload))
	{
		cerr << "fx_compress6lowpan: frame already has a payload field" << endl;
		return false;
	}

	
	// binary of the IPv6 packet
	Buffer buff;
	if (!encode (frame.Get_ipv6(), buff))
		return false;
	buff.SetPosition (0);

	SixLowpanHC& hc = frame.Get_hc();

	// first compressed header is IPv6
	int next_header = NEXT_HEADER_IPV6;
	
	while (next_header >= 0)
	{
		Variable* new_hch = NULL;
		int id = -1;

		switch (next_header)
		{
		case NEXT_HEADER_IPV6:
			id = SixLowpanHCHeader::id_iphc;
			new_hch = compress_iphc (buff, next_header, frame);
			break;
		case NEXT_HEADER_UDP:
			id = SixLowpanHCHeader::id_udp;
			new_hch = compress_nhc_udp (buff, next_header);
			break;
		default:
			cerr << "fx_compress6lowpan: unknown compressed next header type: " << next_header << endl;
		}

		if (new_hch == NULL)
			return false;

		SixLowpanHCHeader* hdr = new SixLowpanHCHeader;
		hdr->SetValue (id, new_hch);
		hc.AppendField (hdr);
	}

	// payload
	frame.Get_payload().Decode (buff);

	// discard the original IPv6 packet
	frame.SetField (frame.id_dispatch, new Undef());
	frame.SetField (frame.id_ipv6, new Undef());

//	frame.Dump (cerr);

	return encode (frame, ret_val);
}

// A DelayedOperation class used for computing the payload length of an IPv6
// header being decompressed.
//
// This operation is delayed because it is necessary to wait until the full
// frame is decompressed in order to know the length of the payload.
class DelayedOperationIPv6ComputePayload : public DelayedOperation
{
public:
	DelayedOperationIPv6ComputePayload (int hdr_position)
	: mPosition (hdr_position)
	{
	}

	bool operator() (Buffer& buff)
	{
		assert (buff.GetLength() >= mPosition + 40*8);

		UInt16 plen;
		plen.SetValue ((buff.GetLength() - mPosition - 40*8) / 8);
		buff.SetPosition (mPosition + 32);
		plen.Encode (buff);
		return true;
	}
private:
	int	mPosition;
};

// @desc	decompress an IPv6 header
//
// @param buff		buffer into which the decompressed header will be written
// @param iphc		IPHC header to be decompressed
// @param next_header	value of the NextHeader field (given if this header is
//			followed by a NHC header)
// @param frame		reference to 6lowpan frame containing this IPv6 packet
// @param delayed_lifo	DelayedLIFO object that will be used to schedule the
//			computation of the IPv6 Payload Length field
void decompress_iphc (Buffer& buff, SixLowpanIPHC& iphc, uint8_t next_header, SixLowpanFrame& frame, DelayedLIFO& delayed_lifo)
{
	// decompress the IPv6 header
	Ipv6Header ip6;

	ip6.Get_version().SetValue (6);

	// ECN + DSCP + Flow Label 
	uint8_t tc = 0;
	if (iphc.IsPresent (iphc.id_inline_ecn))
		tc |= iphc.Get_inline_ecn().GetValue() << 6;
	if (iphc.IsPresent (iphc.id_inline_dscp))
		tc |= iphc.Get_inline_dscp().GetValue();
	ip6.Get_trafficClass().SetValue (tc);

	ip6.Get_flowLabel().SetValue (iphc.IsPresent (iphc.id_inline_fl) ? iphc.Get_inline_fl().GetValue() : 0);

	// hop limit
	uint8_t hl;
	if (iphc.IsPresent (iphc.id_inline_hl)) {
		hl = iphc.Get_inline_hl().GetValue();
	} else {
		static const uint8_t decode_hl[] = {0, 1, 64, 255};
		hl = decode_hl[iphc.Get_hl().GetValue()];
	}
	ip6.Get_hopLimit().SetValue (hl);
	
	// Source Address
	decompress_unicast (	ip6.Get_sourceAddress(),
				frame.Get_hwSrc(),
				iphc.Get_sam(),
				iphc.Get_sac(),
				iphc.Get_inline_src());

	// Destination Address
	if (!iphc.Get_m().GetValue()) {
		// unicast
		decompress_unicast (	ip6.Get_destinationAddress(),
					frame.Get_hwDst(),
					iphc.Get_dam(),
					iphc.Get_dac(),
					iphc.Get_inline_dst());
	} else {
		// multicast
		uint8_t addr[16];
		memset (addr, 0, 16);
		addr[0] = 0xff;

		const uint8_t* inl = iphc.Get_inline_dst().GetValueBin();

		// only Stateless is supported
		switch (iphc.Get_dam().GetValue())
		{
		case 0:	// 128 bits
			memcpy (addr, inl, 16);
			break;
		case 1: // 48 bits
			addr[1] = inl[0];
			memcpy (addr+11, inl+1, 5);
			break;
		case 2:	// 32 bits
			addr[1] = inl[0];
			memcpy (addr+13, inl+1, 3);
			break;
		case 3: // 8 bits
			addr[1] = 2;
			addr[15] = inl[0];
			break;
		}
		ip6.Get_destinationAddress().SetValueBin (addr, 128);

	}
	// next header
	ip6.Get_nextHeader().SetValue (
		iphc.IsPresent (iphc.id_inline_nh)
			? iphc.Get_inline_nh().GetValue()
			: next_header
	);

	// payload length
	delayed_lifo.push_back (new DelayedOperationIPv6ComputePayload(buff.GetPosition()));

	ip6.Encode (buff);
}

// @desc	decompress a compressed UDP header
//
// @param buff		buffer into which the decompressed header will be written
// @param uhc		NHC_UDP header to be decompressed
// @param payload	payload of the UDP message
void decompress_nhc_udp (Buffer& buff, SixLowpanNHCUDP& uhc, Octetstring& payload)
{
	UInt16 u16;

	// src & dst ports
	switch (uhc.Get_p().GetValue())
	{
	case 0: // 16/16
		u16.SetValueBin (uhc.Get_inline_sport());
		u16.Encode (buff);
		u16.SetValueBin (uhc.Get_inline_dport());
		u16.Encode (buff);
		break;
	case 1: // 16/8
		u16.SetValueBin (uhc.Get_inline_sport());
		u16.Encode (buff);

		u16.SetValue (0xf000 + *(uint8_t*)uhc.Get_inline_dport().GetValueBin());
		u16.Encode (buff);
		break;
	case 2: // 8/16
		u16.SetValue (0xf000 + *(uint8_t*)uhc.Get_inline_sport().GetValueBin());
		u16.Encode (buff);

		u16.SetValueBin (uhc.Get_inline_dport());
		u16.Encode (buff);
		break;
	case 3: // 4/4
		u16.SetValue (0xf0b0 + (*(uint8_t*)uhc.Get_inline_sport().GetValueBin() >> 4));
		u16.Encode (buff);
		u16.SetValue (0xf0b0 + (*(uint8_t*)uhc.Get_inline_dport().GetValueBin() >> 4));
		u16.Encode (buff);
		break;
	}

	// message length
	u16.SetValue (payload.GetLength()/8 + 8);
	u16.Encode (buff);

	// checksum
	if (!uhc.Get_c().GetValue()) {
		// inline
		uhc.Get_inline_checksum().Encode (buff);
	} else {
		// elided
		u16.SetValue (0xffff);
		u16.Encode (buff);
	}
}

// @desc	external function fx_decompress6lowpan (defined in Ats6lowpan)
//
// @param 0	a compressed SixLowpanFrame
//
// @return	an uncompressed SixLowpanFrame
bool fx_decompress6lowpan (ParameterList& par_list, Bitstring& ret_val)
{
	SixLowpanFrame frame;
	if (!decode (const_cast<Bitstring&> (par_list[0]), frame))
		return false;

	if (!frame.IsPresent (frame.id_hc))
	{
		// uncompressed frame 
		// -> do nothing
		ret_val.SetValueBin (par_list[0]);
		return true;
	}

	if ((frame.IsPresent (frame.id_dispatch)) |
	    (frame.IsPresent (frame.id_ipv6)))
	{
		cerr << "fx_decompress6lowpan: compressed frame also contains an ipv6 packet" << endl;
		return false;
	}


	if (!frame.IsPresent (frame.id_payload))
	{
		cerr << "fx_decompress6lowpan: compressed frame does not contain a payload field" << endl;
		return false;
	}

	Buffer buff;

	// dispatch (uncompressed IPv6)
	buff.SetValueBin ("\x41", 8);
	buff.SetPosition (8);

	// lifo for delayed operations
	DelayedLIFO delayed_lifo;
	
	SixLowpanHC& hc = frame.Get_hc();
	for (int i=0 ; i<hc.GetSize() ; i++)
	{
		// guess the type of the next compressed header (if any)
		uint8_t next_header = 255;
		if ((i+1) < hc.GetSize()) 
		{
			SixLowpanHCHeader& next_hch = hc.GetField (i+1);
			switch (next_hch.GetChosenId())
			{
			case SixLowpanHCHeader::id_iphc:
				next_header = NEXT_HEADER_IPV6;
				break;
			case SixLowpanHCHeader::id_udp:
				next_header = NEXT_HEADER_UDP;
				break;
			}
		}

		// decompress the actual header
		SixLowpanHCHeader& hch = hc.GetField (i);
		switch (hch.GetChosenId())
		{
		case SixLowpanHCHeader::id_iphc:
			decompress_iphc (buff, hch.Get_iphc(), next_header, frame, delayed_lifo);
			break;
		case SixLowpanHCHeader::id_udp:
			decompress_nhc_udp (buff, hch.Get_udp(), frame.Get_payload());
			break;
		default:
			cerr << "fx_decompress6lowpan: unsupported compressed next header: "
			     << hch.GetChosenFieldName() << endl;
			return false;
		}
	}

	// Append the IPv6 payload
	frame.Get_payload().Encode (buff);

	// discard the IPHC header
	frame.SetField (frame.id_hc, new Undef());

	// flush the delayed lifo
	while (!delayed_lifo.empty())
	{
		buff.SetPosition(0);
		if (! (*delayed_lifo.back()) (buff))
			return false;
		delayed_lifo.pop_back();
	}

	// replace the payload w/ the decoded buffer
	buff.SetPosition(0);
	frame.Get_payload().Decode (buff);

	bool result = encode (frame, ret_val);
	if (!result)
		return false;

/*
	static int nb = 0;
	Buffer newbuf;
	newbuf.SetValueBin (ret_val);
	if (!decode (newbuf, frame))
		return false;
	cout << "------------------------------------------------------------------" << endl;
	cout << "Frame #" << nb++ << endl << endl;
	frame.Dump (cout);
	cout << endl;
	frame.DumpInline (cout);
	cout << endl;
*/	
	return true;
}

namespace t3devlib {

	// @desc	Initialisation function for the Platform Adapter provided by T3DevKit
	//		-> will register all the external functions provided by this module
	void PAInit()
	{
		ExternalFunction::Register ("LibIpv6_ExternalFunctions", "fx_payloadLength", &fx_payloadLength);
		ExternalFunction::Register ("LibIpv6_ExternalFunctions", "fx_icmpv6Checksum", &fx_icmpv6Checksum);
		ExternalFunction::Register ("LibIpv6_ExternalFunctions", "fx_encodeMessage", &fx_encodeMessage);

		ExternalFunction::Register ("Ats6lowpan", "fx_compress6lowpan", &fx_compress6lowpan);
		ExternalFunction::Register ("Ats6lowpan", "fx_decompress6lowpan", &fx_decompress6lowpan);
	}
	
	// @desc	Reset function for the Platform Adapter provided by T3DevKit
	void PAReset()
	{
	}
}

/*
 * Local variables:
 * mode: c++
 * c-basic-offset: 8
 * End:
 */
