#ifndef SixLowpanPort_h
#define SixLowpanPort_h 1

// 6lowpan System Adapter implementation
//
// Please read the "Implementing Adapters" section of the T3DevKit
// documentation in order to understand how this SA is designed.
//
// --->  http://t3devkit.gforge.inria.fr/doc/userref/

#define REMOTE_IP		"127.0.0.1"
#define REMOTE_PORT		13000

#include <t3devlib/t3devlib.h>
#include <boost/asio.hpp>
#include <boost/thread.hpp>

	
class SixLowpanPort : public t3devlib::Port
{
public:
	// @desc	constructor (to be called by T3DevKit's SA)
	SixLowpanPort (t3devlib::PortId& id);

	// @desc	destructor
	~SixLowpanPort();
		
protected:
	// @desc	Map the SixLowpanPort to a TTCN-3 port.
	//
	//		This function is called when the Test Executable
	//		requests to map this system port to a component port.
	//		Then the port will wait for a configuration frame.
	//
	// @note	This port can mapped to only 1 TTCN-3 component at once (this
	//  		is a design constraint)
	//
	// @return	true if successful
	bool Map (const t3devlib::PortId& port_id);


	// @desc	Unmap the SixLowpanPort from a TTCN-3 port.
	//
	//		This function is called when the Test Executable
	//		requests to unmap this system port to a component port.
	//
	// @note	This function will close the listening UDP socket
	// 		(if a configuration frame has been sent)
	//
	// @return	true if successful
	bool Unmap (const t3devlib::PortId& port_id);


	// @desc	Send a frame to this SixLowpanPort
	//
	//		This function is called when the Test Executable
	//		sends a message to this system port.
	//
	// @note	The first message to be sent through this port must be
	//		a SixLowpanPortConfig record holding the configuration
	//		informations for this port. Then, subsequent messages
	//		must be SixLowpanFrame records.
	//
	// @return	true if successful
	bool Send (const t3devlib::ComponentId& from, const t3devlib::Bitstring& msg);
		
		
private:
	// @desc	Tell the Boost ASIO service to start receiving
	//		asynchronously messages from our socket.
	void StartReceive();

	// @desc	Callback function called by the Boost ASIO service
	// 		when a message has been received from the socket
	void HandleReceive (const boost::system::error_code& error, std::size_t nb);

	// @desc	UDP socket to communicate with the PHY adapter
	boost::asio::ip::udp::socket		mSocket;

	// @desc	Pointer to the TTCN-3 component port mapped to this port
	// 		(NULL if unmapped)
	const t3devlib::PortId*			mMappedPort;

	// @desc	Buffer for storing asynchronous messages received by Boost ASIO
	boost::array<char, 2048>		mBuffer;

	// @desc	Error status of the Boost ASIO receiving process
	// 		-1  ->  the process is still running
	// 		 0  ->  the process aborted because of an error
	// 		 1  ->  the process is successfully completed
	int					mReceiveResult;

	// @desc	Boolean flag indicating if this port is configured (see .Send())
	bool					mConfigured;

	// @desc	IEEE 802.15.4 PAN ID used by this port
	uint16_t				mPanId;

	// @desc	mutex protecting state changes on this port
	boost::mutex				mMutex;

	// @desc	condition for reporting the completion of the receive process
	boost::condition_variable		mCondition;
};



#endif
